package cn.com.ty.lift.common.sql;

import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.mapping.BoundSql;
import org.apache.ibatis.mapping.MappedStatement;
import org.apache.ibatis.plugin.Invocation;
import org.apache.ibatis.reflection.MetaObject;
import org.apache.ibatis.reflection.SystemMetaObject;

import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.sql.Statement;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * SQL解析
 *
 * @author wcz
 * @since 2020/2/27
 */
@Slf4j
public abstract class AbstractSqlParser {

    private static final String NEWLINE                        = "\n";
    private static final String SPACE                          = " ";
    private static final String DruidPooledPreparedStatement   = "com.alibaba.druid.pool.DruidPooledPreparedStatement";
    private static final String T4CPreparedStatement           = "oracle.jdbc.driver.T4CPreparedStatement";
    private static final String OraclePreparedStatementWrapper = "oracle.jdbc.driver.OraclePreparedStatementWrapper";
    private static final String h_statement                    = "h.statement";
    private static final String stmt_statement                 = "stmt.statement";
    private static final String delegate                       = "delegate";
    private static final String getSql                         = "getSql";
    private static final String getOriginalSql                 = "getOriginalSql";
    private              Method oracleGetOriginalSqlMethod;
    private              Method druidGetSQLMethod;

    private static final SqlFormatter SQL_FORMATTER             = new SqlFormatter();
    private static final String       h_target                  = "h.target";
    /**
     * MD5
     */
    private static final String       MD5                       = "MD5";
    /**
     * Eight-bit UCS Transformation Format
     */
    private static final  Charset      UTF_8                     = Charset.forName("UTF-8");
    private static final  String       DELEGATE_MAPPED_STATEMENT = "delegate.mappedStatement";
    private static final  String       DELEGATE_BOUND_SQL        = "delegate.boundSql";

    private Statement hStatement(Invocation invocation) {
        Statement statement;
        Object firstArg = invocation.getArgs()[0];
        if (Proxy.isProxyClass(firstArg.getClass())) {
            statement = (Statement) SystemMetaObject.forObject(firstArg).getValue(h_statement);
        } else {
            statement = (Statement) firstArg;
        }
        return statement;
    }

    private Statement stmtStatement(MetaObject metaObject) {
        Statement statement = null;
        try {
            statement = (Statement) metaObject.getValue(stmt_statement);
        } catch (Exception e) {
            // do nothing
            log.warn("Get stmt.statement from MetaObject fail.");
        }
        return statement;
    }

    private Statement delegateStatement(MetaObject metaObject) {
        Statement statement = null;
        if (metaObject.hasGetter(delegate)) {
            //Hikari
            try {
                statement = (Statement) metaObject.getValue(delegate);
            } catch (Exception ignored) {
                // do nothing
                log.warn("Get delegate statement from MetaObject fail.");
            }
        }
        return statement;
    }

    private String originalSqlInDruidPooled(Statement statement) {
        String originalSql = null;
        try {
            if (isNull(druidGetSQLMethod)) {
                Class<?> clazz = Class.forName(DruidPooledPreparedStatement);
                druidGetSQLMethod = clazz.getMethod(getSql);
                druidGetSQLMethod.setAccessible(true);
            }
            Object stmtSql = druidGetSQLMethod.invoke(statement);
            if (isString(stmtSql)) {
                originalSql = (String) stmtSql;
            }
        } catch (Exception e) {
            log.error("Get original Sql from DruidPooledPreparedStatement fail.", e);
        }
        return originalSql;
    }

    private String originalSqlInT4COrOracleWrapper(Statement statement) {
        String originalSql = null;
        try {
            if(isNull(oracleGetOriginalSqlMethod)){
                Class<?> clazz = Class.forName(statement.getClass().getName());
                oracleGetOriginalSqlMethod = getMethodRegular(clazz, getOriginalSql);
            }
            if(notNull(oracleGetOriginalSqlMethod)){
                //OraclePreparedStatementWrapper is not a public class, need set this.
                oracleGetOriginalSqlMethod.setAccessible(true);
                Object stmtSql = oracleGetOriginalSqlMethod.invoke(statement);
                if (isString(stmtSql)) {
                    originalSql = (String) stmtSql;
                }
            }
        } catch (Exception e) {
            //ignore
            log.warn("Get original Sql from T4CPreparedStatement | OraclePreparedStatementWrapper fail.");
        }
        return originalSql;
    }

    public String originalSql(Invocation invocation) {
        Statement statement = hStatement(invocation);

        MetaObject stmtMetaObject = SystemMetaObject.forObject(statement);

        statement = stmtStatement(stmtMetaObject);
        statement = delegateStatement(stmtMetaObject);

        String originalSql = null;
        String stmtClassName = statement.getClass().getName();
        if (DruidPooledPreparedStatement.equals(stmtClassName)) {
            originalSql = originalSqlInDruidPooled(statement);
        } else if (T4CPreparedStatement.equals(stmtClassName) || OraclePreparedStatementWrapper.equals(stmtClassName)) {
            originalSql = originalSqlInT4COrOracleWrapper(statement);
        }
        if (isNull(originalSql)) {
            originalSql = statement.toString();
        }
        return originalSql;
    }

    /**
     * 获取此方法名的具体 Method
     *
     * @param clazz      class 对象
     * @param methodName 方法名
     * @return 方法
     */
    private Method getMethodRegular(Class<?> clazz, String methodName) {
        if (Object.class.equals(clazz)) {
            return null;
        }
        for (Method method : clazz.getDeclaredMethods()) {
            if (method.getName().equals(methodName)) {
                return method;
            }
        }
        return getMethodRegular(clazz.getSuperclass(), methodName);
    }

    /**
     * 获取sql语句开头部分
     *
     * @param sql ignore
     * @return ignore
     */
    private int indexOfSqlStart(String sql) {
        String upperCaseSql = sql.toUpperCase();
        Set<Integer> set = new HashSet<>();
        set.add(upperCaseSql.indexOf("SELECT "));
        set.add(upperCaseSql.indexOf("UPDATE "));
        set.add(upperCaseSql.indexOf("INSERT "));
        set.add(upperCaseSql.indexOf("DELETE "));
        set.remove(-1);
        if (SqlUtils.isEmpty(set)) {
            return -1;
        }
        List<Integer> list = new ArrayList<>(set);
        list.sort(Comparator.naturalOrder());
        return list.get(0);
    }

    /**
     * 格式sql
     *
     * @param boundSql
     * @param format
     * @return
     */
    public String sqlFormat(String boundSql, boolean format) {
        if (format) {
            try {
                return SQL_FORMATTER.format(boundSql);
            } catch (Exception ignored) {
            }
        }
        return boundSql;
    }

    public boolean isNull(Object object){
        return null == object;
    }

    public boolean notNull(Object object){
        return null != object;
    }

    public boolean isString(Object object){
        return (object instanceof String);
    }
    /**
     * 判断字符串是否为空
     *
     * @param cs 需要判断字符串
     * @return 判断结果
     */
    public boolean isEmpty(final CharSequence cs) {
        int strLen;
        if (isNull(cs) || (strLen = cs.length()) == 0) {
            return true;
        }
        for (int i = 0; i < strLen; i++) {
            if (!Character.isWhitespace(cs.charAt(i))) {
                return false;
            }
        }
        return true;
    }

    /**
     * 判断字符串是否不为空
     *
     * @param cs 需要判断字符串
     * @return 判断结果
     */
    public boolean isNotEmpty(final CharSequence cs) {
        return !isEmpty(cs);
    }

    /**
     * 校验集合是否为空
     *
     * @param coll 入参
     * @return boolean
     */
    public boolean isEmpty(Collection<?> coll) {
        return (coll == null || coll.isEmpty());
    }

    /**
     * 判断是不是数字
     *
     * @param cs 输入
     * @return boolean
     */
    public boolean isNumeric(final CharSequence cs) {
        if (isEmpty(cs)) {
            return false;
        }
        Pattern pattern = Pattern.compile("[0-9]+");
        Matcher isNum = pattern.matcher(cs);
        return isNum.matches();
    }

    /**
     * 判断是否为boolean类型的字符串
     *
     * @param cs 输入
     * @return boolean
     */
    public boolean isBoolean(final CharSequence cs) {
        if (isEmpty(cs)) {
            return false;
        }
        String input = cs.toString().toLowerCase();
        return "true".equals(input) || "false".equals(input);
    }

    /**
     * 获得真正的处理对象,可能多层代理.
     */
    @SuppressWarnings("unchecked")
    public <T> T realTarget(Object target) {
        if (Proxy.isProxyClass(target.getClass())) {
            MetaObject metaObject = SystemMetaObject.forObject(target);
            return realTarget(metaObject.getValue(h_target));
        }
        return (T) target;
    }

    /**
     * MD5 Base64 加密
     *
     * @param str 待加密的字符串
     * @return 加密后的字符串
     */
    public String md5Base64(String str) {
        //确定计算方法
        try {
            MessageDigest md5 = MessageDigest.getInstance(MD5);
            //加密后的字符串
            byte[] src = md5.digest(str.getBytes(UTF_8));
            return Base64.getEncoder().encodeToString(src);
        } catch (Exception e) {
            throw new SqlAnalysisException(e);
        }
    }

    /**
     * 获取当前执行 MappedStatement
     *
     * @param metaObject 元对象
     */
    public MappedStatement getMappedStatement(MetaObject metaObject) {
        return (MappedStatement) metaObject.getValue(DELEGATE_MAPPED_STATEMENT);
    }

    public BoundSql getBoundSql(MetaObject metaObject) {
        return (BoundSql) metaObject.getValue(DELEGATE_BOUND_SQL);
    }
}
